<footer class="post-info">
        <abbr class="published" title="{{ article.date.isoformat() }}">
                {{ article.locale_date }}
        </abbr>

        {% if article.author %}
        <address class="vcard author">
                By <a class="url fn" href="{{ SITEURL }}/author/{{ article.author }}.html">{{ article.author }}</a>
        </address>
        {% endif %}
<p>In <a href="{{ SITEURL }}/category/{{ article.category }}.html">{{ article.category }}</a>. {% if PDF_PROCESSOR %}<a href="{{ SITEURL }}/pdf/{{ article.slug }}.pdf">get the pdf</a>{% endif %}</p>
{% include 'taglist.html' %}
{% include 'translations.html' %}
</footer><!-- /.post-info -->
