<!DOCTYPE html>
<html lang="en">
<head>
        <title>{% block title %}{{ SITENAME }}{%endblock%}</title>
        <meta charset="utf-8" />
        <link rel="stylesheet" href="{{ SITEURL }}/theme/css/{{ CSS_FILE }}" type="text/css" />
        <link href="{{ SITEURL }}/{{ FEED }}" type="application/atom+xml" rel="alternate" title="{{ SITENAME }} ATOM Feed" />
        {% if FEED_RSS %}
        <link href="{{ SITEURL }}/{{ FEED_RSS }}" type="application/atom+xml" rel="alternate" title="{{ SITENAME }} RSS Feed" />
        {% endif %}

        <!--[if IE]>
                <script src="http://html5shiv.googlecode.com/svn/trunk/html5.js"></script><![endif]-->

        <!--[if lte IE 7]>
                <link rel="stylesheet" type="text/css" media="all" href="{{ SITEURL }}/css/ie.css"/>
                <script src="{{ SITEURL }}/js/IE8.js" type="text/javascript"></script><![endif]-->

        <!--[if lt IE 7]>
                <link rel="stylesheet" type="text/css" media="all" href="{{ SITEURL }}/css/ie6.css"/><![endif]-->

</head>

<body id="index" class="home">
{% include 'github.html' %}
        <header id="banner" class="body">
                <h1><a href="{{ SITEURL }}">{{ SITENAME }} {% if SITESUBTITLE %} <strong>{{ SITESUBTITLE }}</strong>{% endif %}</a></h1>
                <nav><ul>
                {% for title, link in MENUITEMS %}
                    <li><a href="{{ link }}">{{ title }}</a></li>
                {% endfor %}
                {% if DISPLAY_PAGES_ON_MENU %}
                {% for page in PAGES %}
                    <li><a href="{{ SITEURL }}/pages/{{ page.url }}">{{ page.title }}</a></li>
                {% endfor %}
                {% endif %}
                {% for cat, null in categories %}
                    <li {% if cat == category %}class="active"{% endif %}><a href="{{ SITEURL }}/category/{{ cat }}.html">{{ cat }}</a></li>
                {% endfor %}
                </ul></nav>
        </header><!-- /#banner -->
        {% block content %}
        {% endblock %}
        <section id="extras" class="body">
        {% if LINKS %}
                <div class="blogroll">
                        <h2>blogroll</h2>
                        <ul>
                        {% for name, link in LINKS %}
                            <li><a href="{{ link }}">{{ name }}</a></li>
                        {% endfor %}
                        </ul>
                </div><!-- /.blogroll -->
        {% endif %}
        {% if SOCIAL %}
                <div class="social">
                        <h2>social</h2>
                        <ul>
                            <li><a href="{{ SITEURL }}/{{ FEED }}" rel="alternate">atom feed</a></li>
                            {% if FEED_RSS %}
                            <li><a href="{{ SITEURL }}/{{ FEED_RSS }}" rel="alternate">rss feed</a></li>
                            {% endif %}

                        {% for name, link in SOCIAL %}
                            <li><a href="{{ link }}">{{ name }}</a></li>
                        {% endfor %}
                        </ul>
                </div><!-- /.social -->
        {% endif %}
        </section><!-- /#extras -->

        <footer id="contentinfo" class="body">
                <address id="about" class="vcard body">
                Proudly powered by <a href="http://alexis.notmyidea.org/pelican/">pelican</a>, which takes great advantages of <a href="http://python.org">python</a>.
                </address><!-- /#about -->

                <p>The theme is by <a href="http://coding.smashingmagazine.com/2009/08/04/designing-a-html-5-layout-from-scratch/">Smashing Magazine</a>, thanks!</p>
        </footer><!-- /#contentinfo -->

{% include 'analytics.html' %}
{% include 'piwik.html' %}
{% include 'disqus_script.html' %}
</body>
</html>
